#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>


/*
  Die Leistung vom Code hängt von vielen Faktoren ab.
  Hardware-Eigenschaften, Prozessor, Speicher.
  Ein gewisses Grundverständnis sollte man sich aneignen.
  Befehlszeiger (Instruction pointer), Pipeline, Register, Speicher & Cache sind wichtige Grundbegriffe.

  http://tams-www.informatik.uni-hamburg.de/applets/baukasten/DA/VNR_Einleitung.html
  http://de.wikipedia.org/wiki/Befehlsz%C3%A4hler
  http://de.wikipedia.org/wiki/Pipeline_%28Prozessor%29

  http://de.wikipedia.org/wiki/Intel-Nehalem-Mikroarchitektur

  Ein sehr nettes Dokument zum Thema:
  http://sc.tamu.edu/systems/eos/nehalem.pdf

  Technische Informatik ist hierfür hilfreich.

  Leistungsoptimierung ist anstrengend.
  Das wichtigste ist sich auf die relevanten Bereiche zu konzentrieren.

  Man kann Leistung auch nachträglich messen mit den Performance Countern die in modernen CPUs verbaut sind.
  Somit kann man die Instruktionen pro Sekunde, Speicherbandbreite und Anzahl der Fließkommaoperationen erfassen.

  Probiert auf dem Cluster einmal auf einem der Knoten aus:
  module load likwid
  likwid-perfCtr -c 2  likwid-pin -c 2 <BEFEHL>


  Um sich einen Überblick über den Codebereich zu verschaffen, welcher am meisten/längsten aufgerufen wird, probiert einmal den Code mittels Profiling zu kompilieren:
  gcc -pg 25*.c -o 25-leistungsbewertung -Wall -std=c99

  Nach einem Programmlauf könnt ihr die Ergebnisse mittels gprof ausgeben lassen.
  time ./25-leistungsbewertung
  gprof <Dateiname>

  Das Programm time misst noch die Laufzeit eures Programms.

  Der Compiler erzeugt Code, welcher in festen Zeitabständen die Position des Befehlszählers ermittelt,
  die Ergebnisse werden für die Funktionen gespeichert.

  Mehr Information hier:
    http://www.cs.utah.edu/dept/old/texinfo/as/gprof.html

  Die einfachste Variante der Optimierung besteht darin den Compiler den Code optimieren zu lassen.
  Dafür kann der Compiler verschiedene Verfahren anwenden, z.b. überflüssige Instruktionen des Anwenders entfernen, Variablen in Registern behalten anstelle sofort in Speicher zurück zu schreiben, oder Schleifen-inhalt replizieren (anstelle bedingte Sprünge durchzuführen), den bedingte Sprünge leeren potentiell die Pipeline (da wir nicht wissen ob der Sprung durchgeführt wird).

  Der GCC nutzt die Optimierungen mit -O<Optimierungsstufe>, von -O0 (keine Optimierung) über -O1 und -O2 zu
  -O3, letzteres birgt aber in seltenen Fällen die Gefahr das die Semantik des Codes verändert wurde, daher immer aufpassen, ob der Code sich noch so verhält, wie er sollte...

  Debugging und Optimierungen verträgt sich leider oftmals nicht so gut, da der Code u.U. in anderer Reihenfolge ausgeführt wird und die Variablen beim Inspizieren andere Werte aufweisen, da sie noch im Register sich befinden.
  Also VORSICHT, bei Problemen die Optimierungsstufe erst einmal auf -O0 stellen!

  Probiert einmal aus:
  gcc -pg 25*.c -o 25-leistungsbewertung -Wall -std=c99 -O3
  ./25-leistungsbewertung
  gprof ./25-leistungsbewertung
  Welche Funktionen sollte man jetzt händisch optimieren?


  Auch ganz nett kcachegrind, zum Anzeigen von Cachegrind-Ergebnissen.
  valgrind --tool=cachegrind ./25-leistungsbewertung

  Vorsicht: durch Valgrind Tools verlängert sich die Laufzeit!

  Die Ausgabe von Cachegrind, ein Simulationswerkzeug, welches Prozessorarchitektur simuliert, man beachte, dass heutzutage diese Werte gemessen werden können.
  In früheren Zeiten aber nicht, außerdem kann so die Cache-Größe auf zukünftige / vergangene Systeme angepasst werden.
    ==12513== Cachegrind, a cache and branch-prediction profiler
    ==12513== Copyright (C) 2002-2009, and GNU GPL'd, by Nicholas Nethercote et al.
    ==12513== Using Valgrind-3.6.0.SVN-Debian and LibVEX; rerun with -h for copyright info
    ==12513== Command: ./25-leistungsbewertung
    ==12513==
    --12513-- warning: L3 cache detected but ignored
    ==12513==
    ==12513== I   refs:      18,699,765,185
    ==12513== I1  misses:               938
    ==12513== L2i misses:               927
    ==12513== I1  miss rate:           0.00%
    ==12513== L2i miss rate:           0.00%
    ==12513==
    ==12513== D   refs:       7,993,246,608  (5,481,055,815 rd   + 2,512,190,793 wr)
    ==12513== D1  misses:         1,266,866  (      508,713 rd   +       758,153 wr)
    ==12513== L2d misses:         1,266,435  (      508,322 rd   +       758,113 wr)
    ==12513== D1  miss rate:            0.0% (          0.0%     +           0.0%  )
    ==12513== L2d miss rate:            0.0% (          0.0%     +           0.0%  )
    ==12513==
    ==12513== L2 refs:            1,267,804  (      509,651 rd   +       758,153 wr)
    ==12513== L2 misses:          1,267,362  (      509,249 rd   +       758,113 wr)
    ==12513== L2 miss rate:             0.0% (          0.0%     +           0.0%  )
    ==12683== Branches:       3,408,219,428  (3,208,218,624 cond +   200,000,804 ind)
    ==12683== Mispredicts:           47,062  (       47,000 cond +            62 ind)
    ==12683== Mispred rate:             0.0% (          0.0%     +           0.0%   )
    Profiling timer expired

 */



/*
Im folgenden fünf verschiedene Funktionen, welche Funktion wird wohl wieviel % der Zeit beanspruchen?
 */
#define SIZE 200

void matrixAdd(){
  float Matrix[SIZE][SIZE];
  float InputA[SIZE][SIZE];
  float InputB[SIZE][SIZE];

  for(int i = 0; i < SIZE; i++){
    for(int j = 0; j < SIZE; j++){
      InputA[i][j] = i * j;
      InputB[i][j] = i + j;
    }
  }

  for(int i = 0; i < SIZE; i++){
    for(int j = 0; j < SIZE; j++){
      Matrix[i][j] = InputA[i][j] + InputB[i][j];
    }
  }

  //printf("Results %f\n", Matrix[5][1]);
}

void dynamicMemory(){
  for(long i=0; i < 100000000l; i++){
    char * ptr = malloc(1);
    free(ptr);
  }
}

void repeatMatrixAdd(){
  for(int i=0; i < 1000; i++){
    matrixAdd();
  }
}

void io(){
  for(int i=0; i < 5000; i++){
    char buff[255];
    snprintf(buff, 255, "/tmp/a-%d", i);
    int file = open(buff, O_CREAT | O_APPEND, S_IRWXU);
    close(file);
  }
}


void matrixAddRepeatIf(){
  float Matrix[SIZE][SIZE];
  float InputA[SIZE][SIZE];
  float InputB[SIZE][SIZE];


  for (int repeats = 0; repeats < 1000 ; repeats++){

  for(int i = 0; i < SIZE; i++){
    for(int j = 0; j < SIZE; j++){
      if( i == j){
      }else{ // i != j
      InputA[i][j] = i * j;
      InputB[i][j] = i + j;
      }
    }
  }

    for(int i = 0; i < SIZE; i++){
      for(int j = 0; j < SIZE; j++){
	if( i == j){
	}else{ // i != j
	  Matrix[i][j] = InputA[i][j] + InputB[i][j];
	}
      }
    }
  }

  //printf("Results %f\n", Matrix[5][1]);
}

// ein Funktionszeiger
typedef void (* funcPointer)(void);

int main(int argc, char ** argv){
  // Lese von der Eingabe die aufzurufende Methode
  int method;
  if( argc == 1){
    method = 100;
  }else{
    method = atoi(argv[1]);
  }

  // speichere einen Zeiger auf alle Funktionen
  funcPointer funcs[] = {matrixAdd, dynamicMemory, io, repeatMatrixAdd,matrixAddRepeatIf};

  // rufe alle Methoden auf, falls die Methode 100 ist
  if (method == 100){
    for (int i=0; i < 5; i++){
      funcs[i]();
    }
  }else{
    // sonst rufe genau die passende Methode auf
    funcs[method]();
  }

  return 0;
}

